/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const THUMBSDOWN_CSS_SELECTED = "thumbsdown_selected";

const THUMBSDOWN =
{
    TYPE_DIRECT:  1,
    TYPE_REGEXP:  2,
    TYPE_CUSTOM:  3,
    TYPE_WRAPPED: 4,
    
    TYPE_REGEXP_THUMBNAIL: 1,
    TYPE_REGEXP_LINK:      2,
};


var gThumbsDown =
{
    onLoad: function(e)
    {
        // util.js
        this.util = {};
        var loader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
        loader.loadSubScript("chrome://thumbsdown/content/util.js", this.util);
        
        // localization
        this.util.stringBundle = this.util.$("thumbsdown_stringBundle");
        
        // thumbsdown services
        this.scriptMgr = Cc["@thumbsdown.mozdev.org/script-manager;1"].getService(Ci.nsISupports).wrappedJSObject;
        this.scriptRunner = Cc["@thumbsdown.mozdev.org/script-runner;1"].getService(Ci.nsISupports).wrappedJSObject;
        
        // init shortcut keys
        this.util.initKey("thumbsdown_key_toggleStatus",         "extensions.thumbsdown.keys.toggleStatus");
        this.util.initKey("thumbsdown_key_activateAndSelectAll", "extensions.thumbsdown.keys.activateAndSelectAll");
        this.util.initKey("thumbsdown_key_download",             "extensions.thumbsdown.keys.download");
        
        // stylesheet
        var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
        var ios = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var uri = ios.newURI("chrome://thumbsdown/content/thumbsdown.css", null, null);
        if (! sss.sheetRegistered(uri, sss.USER_SHEET))
            sss.loadAndRegisterSheet(uri, sss.USER_SHEET);
        
        // preference observers
        var prefSvc = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService);
        this.prefs = prefSvc.getBranch("extensions.thumbsdown.");
        this.prefs.QueryInterface(Ci.nsIPrefBranch2);
        this.prefs.addObserver("showInStatusBar", this, false);
        
        // listeners
        gBrowser.tabContainer.addEventListener("TabSelect", function(e) { gThumbsDown.onTabSelect(e) }, false);
        gBrowser.tabContainer.addEventListener("TabClose", function(e) { gThumbsDown.onTabClose(e) }, false);
        gBrowser.addEventListener("DOMContentLoaded", function(e) { gThumbsDown.onPageLoad(e); }, false);
        gBrowser.addEventListener("click", function(e) { gThumbsDown.onContentClick(e); }, true);
        gBrowser.addEventListener("mouseup", function(e) { gThumbsDown.onContentMouseUp(e); }, true);
        
        // popup menu listeners
        this.util.$("menu_ToolsPopup").addEventListener("popupshowing", function(e) { gThumbsDown.onToolsMenuShowing(e); }, false);
        this.util.$("contentAreaContextMenu").addEventListener("popupshowing", function(e) { gThumbsDown.onContextMenuShowing(e); }, false);
        
        // init ui
        var ctx = this._getContext(null, true, false);
        this._updateUi(ctx, true);
    },
    
    onUnload: function(e)
    {
        // preference observers
        this.prefs.removeObserver("showInStatusBar", this);
    },
    
    observe: function(subject, topic, data)
    {
        switch (topic)
        {
            case "nsPref:changed":
                this.observePreference(subject, topic, data)
                break;
            default:
                throw new Error("unknown topic: " + topic);
        }
    },
    
    observePreference: function(subject, topic, data)
    {
        switch (data)
        {
            case "showInStatusBar":
                var ctx = this._getContext(null, true, false);
                this._updateUi(ctx, true);
                break;
            default:
                throw new Error("unknown preference: " + data);
        }
    },
    
    onToolsMenuShowing: function(e)
    {
        var show = nsPreferences.getBoolPref("extensions.thumbsdown.showInToolsMenu");
        this.util.$("thumbsdown_toolsMenu_menu").hidden = (! show);
    },
    
    onContextMenuShowing: function(e)
    {
        var ctx = this._getContext(null, false, false);
        if (ctx && ctx.isActive)
        {
            var hijack = nsPreferences.getBoolPref("extensions.thumbsdown.hijackContextMenu");
            if (hijack)
            {
                e.stopPropagation();
                e.preventDefault();
                var contextMenu = this.util.$("thumbsdown_statusBarContextMenu");
                if (contextMenu.openPopup)
                    contextMenu.openPopup(null, "", e.clientX, e.clientY, true, false);
                else
                    contextMenu.showPopup(contextMenu, e.clientX, e.clientY, "popup");
                return;
            }
        }
        var show = nsPreferences.getBoolPref("extensions.thumbsdown.showInContextMenu");
        this.util.$("thumbsdown_contextMenu_menu").hidden = (! show);
    },
    
    onTabSelect: function(e)
    {
        var ctx = this._getContext(null, true, false);
        this._updateUi(ctx, true);
    },
    
    onTabClose: function(e)
    {
        var browser = e.target.linkedBrowser;
        var ctx = this._getContext(browser, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
    },
    
    onPageLoad: function(e)
    {
        var doc = e.originalTarget;
        if (! (doc instanceof HTMLDocument))
            return;
        if (doc.defaultView.frameElement)
            return;
        var browser = this._getBrowserForDocument(doc);
        var ctx = this._getContext(browser, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
        if (browser != gBrowser.selectedBrowser)
            return;
        this._updateUi(ctx);
    },
    
    onContentClick: function(e)
    {
        if (! (e.button == 0))
            return;
        var elem = e.originalTarget;
        if (! ((elem instanceof HTMLImageElement) &&  (elem.parentNode instanceof HTMLAnchorElement)))
            return;
        var ctx = this._getContext(null, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var changed = ctx.toggleThumbnail(elem);
        e.stopPropagation();
        e.preventDefault();
        if (! changed)
            return;
        this._updateUi(ctx);
    },
    
    onContentMouseUp: function(e)
    {
        var ctx = this._getContext(null, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var changed = this._selectFromSelection(ctx);
        if (! changed)
            return;
        this._updateUi(ctx);
    },
    
    _selectFromSelection: function(ctx)
    {
        var selection = document.commandDispatcher.focusedWindow.getSelection();
        if (selection.isCollapsed)
            return false;
        var range = selection.getRangeAt(0);
        var anchestor = range.commonAncestorContainer;
        var images = [];
        for each (var node in this.util.dfs(anchestor, false))
        {
            if ((node instanceof HTMLImageElement) && (node.parentNode instanceof HTMLAnchorElement))
                if (selection.containsNode(node, false))
                    images.push(node);
            if (node == range.endContainer)
                break;
        }
        var changed = ctx.selectThumbnails(images);
        selection.removeAllRanges();
        return changed;
    },
    
    onStatusBarClick: function(e)
    {
        // single click
        if      ((e.button == 0) && (e.detail == 1))
            this.toggleStatus();
        // double click
        else if ((e.button == 0) && (e.detail == 2))
        {
            this.activate(null, false);
            this.selectAll();
        }
        // single middle click
        else if ((e.button == 1) && (e.detail == 1))
            this.download();
    },
    
    toggleStatus: function(browser)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var ctx = this._getContext(browser, false, false);
        if (ctx && ctx.isActive)
            this.deactivate(browser);
        else
            this.activate(browser, true);
    },
    
    activate: function(browser, selectFromSelection)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var doc = browser.contentDocument;
        // TODO: better valid doc check (image is HTMLDocument)
        if (! (doc instanceof HTMLDocument))
            return;
        var ctx = this._getContext(browser, true, true);
        if (ctx.isActive)
            return;
        if (! this._checkSettings())
            return;
        ctx.isActive = true;
        if (selectFromSelection)
            this._selectFromSelection(ctx);
        else
            document.commandDispatcher.focusedWindow.getSelection().removeAllRanges();
        this._updateUi(ctx);
    },
    
    deactivate: function(browser)
    {
        var ctx = this._getContext(browser, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
        this._updateUi(ctx);
    },
    
    selectAll: function(browser)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var ctx = this._getContext(browser, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var images = browser.contentDocument.images;
        var changed = ctx.selectThumbnails(images);
        if (changed)
            this._updateUi(ctx);
    },
    
    download: function(browser)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var ctx = this._getContext(browser, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var docUrl = browser.contentDocument.location.href;
        var urls = [];
        var scriptArgs = [];
        for each (var image in ctx.thumbnails)
        {
            var script = image.thumbsdown;
            switch (script.type)
            {
                case THUMBSDOWN.TYPE_DIRECT:
                    var url = {};
                    url.url = image.parentNode.href;
                    url.referer = docUrl;
                    urls.push(url);
                    break;
                    
                case THUMBSDOWN.TYPE_REGEXP:
                    var url = {};
                    var sourceUrl = ((script.subType == THUMBSDOWN.TYPE_REGEXP_THUMBNAIL) ? image.src : image.parentNode.href);
                    url.url = sourceUrl.replace(script.re_pattern, script.template);
                    url.referer = image.parentNode.href;
                    urls.push(url);
                    break;
                    
                case THUMBSDOWN.TYPE_CUSTOM:
                    var args = {};
                    args.name = script.name;
                    args.code = script.script;
                    args.documentUrl = docUrl;
                    args.thumbnailUrl = image.src;
                    args.linkUrl = image.parentNode.href;
                    scriptArgs.push(args);
                    break;
                    
                case THUMBSDOWN.TYPE_WRAPPED:
                    // TODO: maybe something better than using code templates...?
                    var args = {};
                    args.name = script.name;
                    args.code = this.scriptRunner.getWrappedScriptCode(script.action1, script.action2);
                    args.documentUrl = docUrl;
                    args.thumbnailUrl = image.src;
                    args.linkUrl = image.parentNode.href;
                    scriptArgs.push(args);
                    break;
                    
                default:
                    throw new Error("invalid type: " + script.type);
            }
        }
        var dir = nsPreferences.copyUnicharPref("extensions.thumbsdown.downloadDir");
        if (urls.length > 0)
            this._downloadUrls(urls, dir);
        if (scriptArgs.length > 0)
            this._runScripts(scriptArgs, dir);
        ctx.isActive = false;
        this._updateUi(ctx);
    },
    
    _downloadUrls: function(urls, dir)
    {
        try
        {
            var dmSvc = Cc["@thumbsdown.mozdev.org/download-manager-service;1"].getService(Ci.nsISupports).wrappedJSObject;
            dmSvc.download(urls, dir);
        }
        catch (e)
        {
            var promptSvc = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
            var title = this.util._("error.title");
            var msg = this.util._("error.download", e.message);
            promptSvc.alert(window, title, msg);
        }
    },
    
    _runScripts: function(scriptArgs, dir)
    {
        var callback = this._getScriptCallback(dir);
        scriptArgs.forEach(function(args) { args.callback = callback });
        this.scriptRunner.addScripts(scriptArgs);
        
        if (nsPreferences.getBoolPref("extensions.thumbsdown.scriptrunner.open"))
        {
            var focus = nsPreferences.getBoolPref("extensions.thumbsdown.scriptrunner.focus")
            this.util.openScriptRunnerWindow(focus);
        }
    },
    
    /**
     * Retrieves ThumbsDown Context object for browser.
     *
     * @param browser 
     * browser from where the context is looked up. 
     * If null, defaults to current tab.
     * @param create if true and context doesn't exist, creates a new
     * @param persist if true, persists created context
     *
     * @return context object or null, if it doesn't exist and a new was not created
     */
    _getContext: function(browser, create, persist)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var ctx = browser.thumbsdown;
        if (ctx)
            return ctx;
        if (! create)
            return null;
        ctx = new ThumbsDownContext();
        if (persist)
            browser.thumbsdown = ctx;
        return ctx;
    },
    
    _checkSettings: function()
    {
        var promptSvc = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
        var title = this.util._("error.title");
        
        var dirPath = nsPreferences.copyUnicharPref("extensions.thumbsdown.downloadDir");
        
        try
        {
            var dir = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
            dir.initWithPath(dirPath);
        }
        catch (e)
        {
            dir = null;
        }
        
        // is download directory set and valid file path?
        if ((dirPath.trim().length == 0) || (! dir))
        {
            var msg = this.util._("error.noDir");
            var button0Label = this.util._("openPreferences");
            var flags =
                promptSvc.BUTTON_POS_0 * promptSvc.BUTTON_TITLE_IS_STRING +
                promptSvc.BUTTON_POS_1 * promptSvc.BUTTON_TITLE_CANCEL;
            
            var result = promptSvc.confirmEx(window, title, msg, flags, button0Label, null, null, null, {});
            
            if (result == 0)
                this.util.openPrefWindow("main");
            
            return false;
        }
        
        // does download directory exist?
        var warnMissingFolder = nsPreferences.getBoolPref("extensions.thumbsdown.warnMissingDownloadFolder");
        if (warnMissingFolder && (! (dir.exists() && dir.isDirectory())))
        {
            var msg = this.util._("error.dirDoesNotExist", dirPath);
            var button0Label = this.util._("openPreferences");
            var checkLabel = this.util._("warn.check");
            var check = {value: false};
            var flags =
                promptSvc.BUTTON_POS_0 * promptSvc.BUTTON_TITLE_IS_STRING +
                promptSvc.BUTTON_POS_1 * promptSvc.BUTTON_TITLE_CANCEL;
            
            var result = promptSvc.confirmEx(window, title, msg, flags, button0Label, null, null, checkLabel, check);
            
            if (check.value)
                nsPreferences.setBoolPref("extensions.thumbsdown.warnMissingDownloadFolder", false);
            
            if (result == 0)
                this.util.openPrefWindow("main");
            
            return false;
        }
        
        // any scripts?
        if (this.scriptMgr.scripts.length == 0)
        {
            var msg = this.util._("error.noScripts");
            var button0Label = this.util._("openHomePage");
            var flags =
                promptSvc.BUTTON_POS_0 * promptSvc.BUTTON_TITLE_IS_STRING +
                promptSvc.BUTTON_POS_1 * promptSvc.BUTTON_TITLE_CANCEL;
            
            var result = promptSvc.confirmEx(window, title, msg, flags, button0Label, null, null, null, {});
            
            if (result == 0)
                this.util.openHomePage();
            
            return false;
        }
        
        return true;
    },
    
    _getScriptCallback: function(dir)
    {
        return function(script)
        {
            var dmSvc = Cc["@thumbsdown.mozdev.org/download-manager-service;1"].getService(Ci.nsISupports).wrappedJSObject;
            var urls = [{url: url, referer: script.referer} for each (url in script.result)];
            dmSvc.download(urls, dir);
        };
    },
    
    _updateUi: function(ctx, force)
    {
        var $ = this.util.$;
        
        // is update neccessary
        var currentContext = this._getContext(null, false, false);
        var isUiActive = ($("thumbsdown_statusBarPanel").getAttribute("active") == "true");
        if (! (force || ((ctx.isActive || isUiActive) && (ctx == currentContext))))
            return;
        
        // status bar visibility
        var showStatusBar = nsPreferences.getBoolPref("extensions.thumbsdown.showInStatusBar");
        showStatusBar = showStatusBar || ctx.isActive;
        
        // status bar
        $("thumbsdown_statusBarPanel").hidden = (! showStatusBar);
        if (showStatusBar)
        {
            $("thumbsdown_statusBarPanel").setAttribute("active", ctx.isActive);
            $("thumbsdown_statusBarTooltip").setAttribute("active", ctx.isActive);
            if (ctx.isActive)
                $("thumbsdown_statusBarLabel").setAttribute("value", "" + ctx.thumbnails.length);
        }
        
        // commands
        $("thumbsdown_cmd_activate").setAttribute("disabled", ctx.isActive);
        $("thumbsdown_cmd_deactivate").setAttribute("disabled", (! ctx.isActive));
        $("thumbsdown_cmd_selectAll").setAttribute("disabled", (! ctx.isActive));
        $("thumbsdown_cmd_download").setAttribute("disabled", (! ctx.isActive));
    },
    
    _getBrowserForDocument: function(doc)
    {
        var browsers = gBrowser.browsers;
        for (var i = 0; i < browsers.length; i++)
            if (browsers[i].contentDocument == doc)
                return browsers[i];
        throw new Error("browser not found");
    }
}


function ThumbsDownContext()
{
    this._isActive = false;
    this.thumbnails = [];
}

ThumbsDownContext.prototype =
{
    get isActive() { return this._isActive; },
    set isActive(value)
    {
        if (value)
            this._isActive = true;
        else
        {
            for each (var image in this.thumbnails)
                this._markThumbnail(image, false);
            this._isActive = false;
            this.thumbnails = [];
        }
    },
    
    toggleThumbnail: function(image)
    {
        var index = this.thumbnails.indexOf(image);
        var selected = (index != -1);
        if (selected)
            this.deselectThumbnail(index);
        else
            selected = this.selectThumbnails([image]);
        return selected;
    },
    
    selectThumbnails: function(images)
    {
        var selected = [];
        // images can be HTMLCollection (document.images) -> using normal for loop
        for (var i = 0; i < images.length; i++)
        {
            var image = images[i];
            if (this.thumbnails.indexOf(image) == -1)
            {
                var script = this._isThumbnail(image);
                if (script)
                {
                    selected.push(image);
                    this._markThumbnail(image, true);
                    image.thumbsdown = script;
                }
            }
        }
        this.thumbnails = this.thumbnails.concat(selected);
        return (selected.length > 0);
    },
    
    deselectThumbnail: function(index)
    {
        var image = this.thumbnails.splice(index, 1)[0];
        this._markThumbnail(image, false);
    },
    
    _isThumbnail: function(image)
    {
        var anchor = image.parentNode;
        if (anchor && anchor.href && (anchor instanceof HTMLAnchorElement))
            return gThumbsDown.scriptMgr.lookup(anchor.href, image.src);
        return null;
    },
    
    _markThumbnail: function(image, selected)
    {
        var classNames = image.className.split(/\s+/);
        if (selected)
            classNames.push(THUMBSDOWN_CSS_SELECTED);
        else
            classNames = [x for each (x in classNames) if (x != THUMBSDOWN_CSS_SELECTED)];
        image.className = classNames.join(' ');
        if (! selected)
            image.thumbsdown = undefined; // delete is not allowed
    }
}


window.addEventListener("load", function(e) { gThumbsDown.onLoad(e); }, false);
window.addEventListener("unload", function(e) { gThumbsDown.onUnload(e); }, false);
